from typing import (
    Optional,
    Callable,
    Any
)

class Depends:
    def __init__(
        self, dependency: Optional[Callable[..., Any]] = None, *, use_cache: bool = True
    ):
        self.dependency = dependency
        self.use_cache = use_cache

    def __repr__(self) -> str:
        attr = getattr(self.dependency, "__name__", type(self.dependency).__name__)
        cache = "" if self.use_cache else ", use_cache=False"
        return f"{self.__class__.__name__}({attr}{cache})"


import asyncio
from typing import (
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Sequence,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)
from gazze.exceptions import RequestValidationError
from gazze.datastructures import Default, DefaultPlaceholder
from gazze.types import DecoratedCallable, IncEx
from gazze.responses import JSONResponse
from gazze.routing import GRouter
from gazze.params import Depends
from gazze.exception_handlers import (
    http_exception_handler,
    request_validation_exception_handler,
)
from starlette.routing import Route, Router
from starlette.applications import Starlette
from starlette.datastructures import State
from starlette.middleware import Middleware
from starlette.responses import Response
from starlette.middleware.errors import ServerErrorMiddleware
from starlette.middleware.exceptions import ExceptionMiddleware
from gazze.middleware.asyncexitstack import AsyncExitStackMiddleware
from starlette.types import ASGIApp, Scope, Receive, Send
# from starlette.types import ASGIApp, Lifespan, Receive, Scope, Send
from starlette.requests import Request
from starlette.exceptions import HTTPException
from hypercorn.asyncio import serve

AppType = TypeVar("AppType", bound="Gazze")

class Gazze(Starlette):
    def __init__(
        self: AppType,
        debug: bool = False,
        title: str = "Gazze",
        description: str = "",
        version: str = "0.1.0",
        exception_handlers: Optional[
            Dict[
                Union[int, Type[Exception]],
                Callable[[Request, Any], Coroutine[Any, Any, Response]],
            ]
        ] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> None:
        self.debug = debug
        self.title = title
        self.description = description
        self.version = version
        self.state: State = State()
        self.router: GRouter = GRouter()
        self.default_response_class = default_response_class
        self.response_class = response_class
        self.dependencies = dependencies
        self.dependency_overrides: Dict[Callable[..., Any], Callable[..., Any]] = {}
        self.exception_handlers: Dict[
            Any, Callable[[Request, Any], Union[Response, Awaitable[Response]]]
        ] = ({} if exception_handlers is None else dict(exception_handlers))
        self.exception_handlers.setdefault(HTTPException, http_exception_handler)
        self.exception_handlers.setdefault(
            RequestValidationError, request_validation_exception_handler
        )
        self.user_middleware: List[Middleware] = []
        self.middleware_stack: Union[ASGIApp, None] = None

    def add_route(
        self,
        *,
        path: str,
        endpoint: Callable[..., Coroutine[Any, Any, Response]],
        methods: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> None:
        self.router.add_route(
            path=path, 
            endpoint=endpoint, 
            methods=methods, 
            dependencies=dependencies,
            response_class=response_class
        )

    def build_middleware_stack(self) -> ASGIApp:
        debug = self.debug
        error_handler = None
        exception_handlers = {}

        for key, value in self.exception_handlers.items():
            if key in (500, Exception):
                error_handler = value
            else:
                exception_handlers[key] = value

        middleware = (
            [Middleware(ServerErrorMiddleware, handler=error_handler, debug=debug)]
            + self.user_middleware
            + [
                Middleware(
                    ExceptionMiddleware, handlers=exception_handlers, debug=debug
                ),
                Middleware(AsyncExitStackMiddleware),
            ]
        )

        app = self.router
        for cls, options in reversed(middleware):
            app = cls(app=app, **options)
        return app
    
    def get(
        self,
        path: str,
        *,
        methods: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.get(
            path=path, 
            dependencies=dependencies,
            response_class=response_class
        )

    def put(
        self,
        path: str,
        *,
        methods: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.put(
            path=path, 
            dependencies=dependencies,
            response_class=response_class
        )

    def post(
        self,
        path: str,
        *,
        methods: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.post(
            path=path, 
            dependencies=dependencies,
            response_class=response_class
        )

    def delete(
        self,
        path: str,
        *,
        methods: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.delete(
            path=path, 
            dependencies=dependencies,
            response_class=response_class
        )

    def options(
        self,
        path: str,
        *,
        methods: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.options(
            path=path, 
            dependencies=dependencies,
            response_class=response_class
        )

    def head(
        self,
        path: str,
        *,
        methods: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.head(
            path=path, 
            dependencies=dependencies,
            response_class=response_class
        )

    def patch(
        self,
        path: str,
        *,
        methods: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.patch(
            path=path, 
            dependencies=dependencies,
            response_class=response_class
        )


from typing import (
    Any,
    Callable,
#     Coroutine,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
#     Tuple,
    Type,
    Union,
)
from gazze.utils import get_value_or_default
from starlette.routing import BaseRoute, Route
from gazze.types import DecoratedCallable#, IncEx
from gazze.params import Depends
from gazze.responses import JSONResponse
from starlette.responses import Response
from gazze.datastructures import Default, DefaultPlaceholder
from starlette.routing import Router


class GRouter(Router):

    def __init__(
        self,
        *,
        route_class: Type[Route] = Route,
        dependencies: Optional[Sequence[Depends]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        routes: Optional[List[BaseRoute]] = None,
        redirect_slashes: bool = True,
    ) -> None:
        super().__init__(
            routes=routes,
            redirect_slashes=redirect_slashes,
        )
        self.route_class = route_class
        self.dependencies = list(dependencies or [])
        self.responses = responses or {}
        self.default_response_class = default_response_class

    def add_api_route(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        response_model: Any = Default(None),
        dependencies: Optional[Sequence[Depends]] = None,
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        methods: Optional[Union[Set[str], List[str]]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
        route_class_override: Optional[Type[Route]] = None,
    ) -> None:
        route_class = route_class_override or self.route_class
        responses = responses or {}
        combined_responses = {**self.responses, **responses}
        current_response_class = get_value_or_default(
            response_class, self.default_response_class
        )
        current_dependencies = self.dependencies.copy()
        if dependencies:
            current_dependencies.extend(dependencies)
        route = route_class(
            path,
            endpoint=endpoint,
            methods=methods,
        )
        self.routes.append(route)

    def api_route(
        self,
        path: str,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        methods: Optional[List[str]] = None,
        response_class: Type[Response] = Default(JSONResponse),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        def decorator(func: DecoratedCallable) -> DecoratedCallable:
            self.add_api_route(
                path,
                func,
                dependencies=dependencies,
                responses=responses,
                methods=methods,
                response_class=response_class,
            )
            return func

        return decorator

    def get(
        self,
        path: str,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.api_route(
            path=path, 
            methods=['GET'],
            dependencies=dependencies,
            response_class=response_class
        )
    
    def put(
        self,
        path: str,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.api_route(
            path=path, 
            methods=['PUT'],
            dependencies=dependencies,
            response_class=response_class
        )
    
    def post(
        self,
        path: str,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.api_route(
            path=path, 
            methods=['POST'],
            dependencies=dependencies,
            response_class=response_class
        )
    
    def delete(
        self,
        path: str,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.api_route(
            path=path, 
            methods=['DELETE'],
            dependencies=dependencies,
            response_class=response_class
        )
    
    def options(
        self,
        path: str,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.api_route(
            path=path, 
            methods=['OPTIONS'],
            dependencies=dependencies,
            response_class=response_class
        )

    def head(
        self,
        path: str,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.api_route(
            path=path, 
            methods=['HEAD'],
            dependencies=dependencies,
            response_class=response_class
        )

    def patch(
        self,
        path: str,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.api_route(
            path=path, 
            methods=['PATCH'],
            dependencies=dependencies,
            response_class=response_class
        )


import time
import click

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def title(text: str):
    return f'{bcolors.OKCYAN}gazze-cli{bcolors.ENDC} {text}'

def byell(text: str):
    return f'{bcolors.WARNING}{bcolors.BOLD}{text}{bcolors.ENDC}{bcolors.ENDC}'

def bgree(text: str):
    return f'{bcolors.OKGREEN}{bcolors.BOLD}{text}{bcolors.ENDC}{bcolors.ENDC}'

@click.group()
def main(): pass


import re
from typing import (
    Any,
    Dict,
    Set,
    Union,
    cast,
)
from gazze.datastructures import DefaultPlaceholder, DefaultType

def is_body_allowed_for_status_code(status_code: Union[int, str, None]) -> bool:
    if status_code is None:
        return True
    if status_code in {
        "default",
        "1XX",
        "2XX",
        "3XX",
        "4XX",
        "5XX",
    }:
        return True
    current_status_code = int(status_code)
    return not (current_status_code < 200 or current_status_code in {204, 304})


def get_path_param_names(path: str) -> Set[str]:
    return set(re.findall("{(.*?)}", path))


def deep_dict_update(main_dict: Dict[Any, Any], update_dict: Dict[Any, Any]) -> None:
    for key, value in update_dict.items():
        if (
            key in main_dict
            and isinstance(main_dict[key], dict)
            and isinstance(value, dict)
        ):
            deep_dict_update(main_dict[key], value)
        elif (
            key in main_dict
            and isinstance(main_dict[key], list)
            and isinstance(update_dict[key], list)
        ):
            main_dict[key] = main_dict[key] + update_dict[key]
        else:
            main_dict[key] = value

def get_value_or_default(
    first_item: Union[DefaultPlaceholder, DefaultType],
    *extra_items: Union[DefaultPlaceholder, DefaultType],
) -> Union[DefaultPlaceholder, DefaultType]:
    items = (first_item,) + extra_items
    for item in items:
        if not isinstance(item, DefaultPlaceholder):
            return item
    return first_item


import dataclasses
from pydantic import BaseModel
from collections import defaultdict, deque
import datetime
from decimal import Decimal
from enum import Enum
from pydantic.networks import AnyUrl, NameEmail
from pydantic.types import SecretBytes, SecretStr
from pydantic_core import Url
from pydantic.color import Color
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from pathlib import Path, PurePath
from re import Pattern
from types import GeneratorType
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Type, Union
from uuid import UUID

def _model_dump(
        model: BaseModel, mode: Literal["json", "python"] = "json", **kwargs: Any
    ) -> Any:
        return model.model_dump(mode=mode, **kwargs)
    
def isoformat(o: Union[datetime.date, datetime.time]) -> str:
    return o.isoformat()

def decimal_encoder(dec_value: Decimal) -> Union[int, float]:
    if dec_value.as_tuple().exponent >= 0:  # type: ignore[operator]
        return int(dec_value)
    else:
        return float(dec_value)

ENCODERS_BY_TYPE: Dict[Type[Any], Callable[[Any], Any]] = {
    bytes: lambda o: o.decode(),
    Color: str,
    datetime.date: isoformat,
    datetime.datetime: isoformat,
    datetime.time: isoformat,
    datetime.timedelta: lambda td: td.total_seconds(),
    Decimal: decimal_encoder,
    Enum: lambda o: o.value,
    frozenset: list,
    deque: list,
    GeneratorType: list,
    IPv4Address: str,
    IPv4Interface: str,
    IPv4Network: str,
    IPv6Address: str,
    IPv6Interface: str,
    IPv6Network: str,
    NameEmail: str,
    Path: str,
    Pattern: lambda o: o.pattern,
    SecretBytes: str,
    SecretStr: str,
    set: list,
    UUID: str,
    Url: str,
    AnyUrl: str,
}

def generate_encoders_by_class_tuples(
    type_encoder_map: Dict[Any, Callable[[Any], Any]]
) -> Dict[Callable[[Any], Any], Tuple[Any, ...]]:
    encoders_by_class_tuples: Dict[Callable[[Any], Any], Tuple[Any, ...]] = defaultdict(
        tuple
    )
    for type_, encoder in type_encoder_map.items():
        encoders_by_class_tuples[encoder] += (type_,)
    return encoders_by_class_tuples


encoders_by_class_tuples = generate_encoders_by_class_tuples(ENCODERS_BY_TYPE)

def jsonable_encoder(
    obj: Any,
    include: Optional[Any] = None,
    exclude: Optional[Any] = None,
    by_alias: bool = True,
    exclude_unset: bool = False,
    exclude_defaults: bool = False,
    exclude_none: bool = False,
    custom_encoder: Optional[Dict[Any, Callable[[Any], Any]]] = None,
    sqlalchemy_safe: bool = True,
) -> Any:
    custom_encoder = custom_encoder or {}
    if custom_encoder:
        if type(obj) in custom_encoder:
            return custom_encoder[type(obj)](obj)
        else:
            for encoder_type, encoder_instance in custom_encoder.items():
                if isinstance(obj, encoder_type):
                    return encoder_instance(obj)
    if include is not None and not isinstance(include, (set, dict)):
        include = set(include)
    if exclude is not None and not isinstance(exclude, (set, dict)):
        exclude = set(exclude)
    if isinstance(obj, BaseModel):
        # TODO: remove when deprecating Pydantic v1
        encoders: Dict[Any, Any] = {}
        obj_dict = _model_dump(
            obj,
            mode="json",
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
        )
        if "__root__" in obj_dict:
            obj_dict = obj_dict["__root__"]
        return jsonable_encoder(
            obj_dict,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
            # TODO: remove when deprecating Pydantic v1
            custom_encoder=encoders,
            sqlalchemy_safe=sqlalchemy_safe,
        )
    if dataclasses.is_dataclass(obj):
        obj_dict = dataclasses.asdict(obj)
        return jsonable_encoder(
            obj_dict,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            custom_encoder=custom_encoder,
            sqlalchemy_safe=sqlalchemy_safe,
        )
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, PurePath):
        return str(obj)
    if isinstance(obj, (str, int, float, type(None))):
        return obj
    if isinstance(obj, dict):
        encoded_dict = {}
        allowed_keys = set(obj.keys())
        if include is not None:
            allowed_keys &= set(include)
        if exclude is not None:
            allowed_keys -= set(exclude)
        for key, value in obj.items():
            if (
                (
                    not sqlalchemy_safe
                    or (not isinstance(key, str))
                    or (not key.startswith("_sa"))
                )
                and (value is not None or not exclude_none)
                and key in allowed_keys
            ):
                encoded_key = jsonable_encoder(
                    key,
                    by_alias=by_alias,
                    exclude_unset=exclude_unset,
                    exclude_none=exclude_none,
                    custom_encoder=custom_encoder,
                    sqlalchemy_safe=sqlalchemy_safe,
                )
                encoded_value = jsonable_encoder(
                    value,
                    by_alias=by_alias,
                    exclude_unset=exclude_unset,
                    exclude_none=exclude_none,
                    custom_encoder=custom_encoder,
                    sqlalchemy_safe=sqlalchemy_safe,
                )
                encoded_dict[encoded_key] = encoded_value
        return encoded_dict
    if isinstance(obj, (list, set, frozenset, GeneratorType, tuple, deque)):
        encoded_list = []
        for item in obj:
            encoded_list.append(
                jsonable_encoder(
                    item,
                    include=include,
                    exclude=exclude,
                    by_alias=by_alias,
                    exclude_unset=exclude_unset,
                    exclude_defaults=exclude_defaults,
                    exclude_none=exclude_none,
                    custom_encoder=custom_encoder,
                    sqlalchemy_safe=sqlalchemy_safe,
                )
            )
        return encoded_list

    if type(obj) in ENCODERS_BY_TYPE:
        return ENCODERS_BY_TYPE[type(obj)](obj)
    for encoder, classes_tuple in encoders_by_class_tuples.items():
        if isinstance(obj, classes_tuple):
            return encoder(obj)

    try:
        data = dict(obj)
    except Exception as e:
        errors: List[Exception] = []
        errors.append(e)
        try:
            data = vars(obj)
        except Exception as e:
            errors.append(e)
            raise ValueError(errors) from e
    return jsonable_encoder(
        data,
        include=include,
        exclude=exclude,
        by_alias=by_alias,
        exclude_unset=exclude_unset,
        exclude_defaults=exclude_defaults,
        exclude_none=exclude_none,
        custom_encoder=custom_encoder,
        sqlalchemy_safe=sqlalchemy_safe,
    )